import LRUCache from 'lru-cache';
import { MaybeCancelablePromise } from '@cubejs-backend/shared';
import { BaseDriver, InlineTables } from '@cubejs-backend/base-driver';
import { QueryQueue } from './QueryQueue';
import { CacheDriverInterface } from './cache-driver.interface';
import { DriverFactory, DriverFactoryByDataSource } from './DriverFactory';
import { PreAggregationDescription } from './PreAggregations';
declare type QueryOptions = {
    external?: boolean;
    renewalThreshold?: number;
    updateWindowSeconds?: number;
    renewalThresholdOutsideUpdateWindow?: number;
    incremental?: boolean;
};
export declare type QueryTuple = [sql: string, params: unknown[], options?: QueryOptions];
export declare type QueryWithParams = QueryTuple;
export declare type Query = {
    requestId?: string;
    dataSource: string;
    preAggregations?: PreAggregationDescription[];
    groupedPartitionPreAggregations?: PreAggregationDescription[][];
    preAggregationsLoadCacheByDataSource?: any;
    renewQuery?: boolean;
};
declare type CacheEntry = {
    time: number;
    result: any;
    renewalKey: string;
};
export declare class QueryCache {
    protected readonly redisPrefix: string;
    protected readonly driverFactory: DriverFactoryByDataSource;
    protected readonly logger: any;
    readonly options: {
        refreshKeyRenewalThreshold?: number;
        externalQueueOptions?: any;
        externalDriverFactory?: DriverFactory;
        backgroundRenew?: Boolean;
        queueOptions?: (dataSource: string) => Promise<{
            concurrency: number;
            continueWaitTimeout?: number;
            executionTimeout?: number;
            orphanedTimeout?: number;
            heartBeatInterval?: number;
        }>;
        redisPool?: any;
        continueWaitTimeout?: number;
        cacheAndQueueDriver?: 'redis' | 'memory';
        maxInMemoryCacheEntries?: number;
        skipExternalCacheAndQueue?: boolean;
    };
    protected readonly cacheDriver: CacheDriverInterface;
    protected queue: {
        [dataSource: string]: QueryQueue;
    };
    protected externalQueue: QueryQueue | null;
    protected memoryCache: LRUCache<string, CacheEntry>;
    constructor(redisPrefix: string, driverFactory: DriverFactoryByDataSource, logger: any, options?: {
        refreshKeyRenewalThreshold?: number;
        externalQueueOptions?: any;
        externalDriverFactory?: DriverFactory;
        backgroundRenew?: Boolean;
        queueOptions?: (dataSource: string) => Promise<{
            concurrency: number;
            continueWaitTimeout?: number;
            executionTimeout?: number;
            orphanedTimeout?: number;
            heartBeatInterval?: number;
        }>;
        redisPool?: any;
        continueWaitTimeout?: number;
        cacheAndQueueDriver?: 'redis' | 'memory';
        maxInMemoryCacheEntries?: number;
        skipExternalCacheAndQueue?: boolean;
    });
    /**
     * Force reconcile queue logic to be executed.
     */
    forceReconcile(datasource?: string): Promise<void>;
    cachedQueryResult(queryBody: any, preAggregationsTablesToTempTables: any): Promise<{
        data: any;
        refreshKeyValues: any[];
        lastRefreshTime: Date;
    } | {
        data: any;
        lastRefreshTime?: undefined;
    } | {
        data: any;
        lastRefreshTime: Date;
    }>;
    private getExpireSecs;
    private cacheKeyQueriesFrom;
    static queryCacheKey(queryBody: any): any[];
    protected static replaceAll(replaceThis: any, withThis: any, inThis: any): any;
    static replacePreAggregationTableNames(queryAndParams: QueryWithParams, preAggregationsTablesToTempTables: any): any;
    queryWithRetryAndRelease(query: any, values: any, { priority, cacheKey, external, requestId, dataSource, inlineTables, useCsvQuery, }: {
        priority?: number;
        cacheKey: object;
        external: boolean;
        requestId?: string;
        dataSource: string;
        inlineTables?: InlineTables;
        useCsvQuery?: boolean;
    }): Promise<any>;
    getQueue(dataSource?: string): Promise<QueryQueue>;
    private csvQuery;
    getExternalQueue(): QueryQueue;
    static createQueue(redisPrefix: string, clientFactory: DriverFactory, executeFn: (client: BaseDriver, q: any) => any, options?: Record<string, any>): QueryQueue;
    /**
     * Returns registered queries queues hash table.
     */
    getQueues(): {
        [dataSource: string]: QueryQueue;
    };
    startRenewCycle(query: any, values: any, cacheKeyQueries: any, expireSecs: any, cacheKey: any, renewalThreshold: any, options: {
        requestId?: string;
        skipRefreshKeyWaitForRenew?: boolean;
        external?: boolean;
        dataSource: string;
    }): void;
    renewQuery(query: any, values: any, cacheKeyQueries: any, expireSecs: any, cacheKey: any, renewalThreshold: any, options: {
        requestId?: string;
        skipRefreshKeyWaitForRenew?: boolean;
        external?: boolean;
        dataSource: string;
        useCsvQuery?: boolean;
    }): Promise<{
        data: any;
        refreshKeyValues: any[];
        lastRefreshTime: Date;
    }>;
    loadRefreshKeysFromQuery(query: Query): Promise<any[]>;
    loadRefreshKeys(cacheKeyQueries: QueryWithParams[], expireSecs: number, options: {
        requestId?: string;
        skipRefreshKeyWaitForRenew?: boolean;
        dataSource: string;
    }): Promise<any>[];
    withLock: <T = any>(key: string, ttl: number, callback: () => import("@cubejs-backend/shared").Optional<import("@cubejs-backend/shared").CancelablePromise<T>, "cancel">) => import("@cubejs-backend/shared").CancelablePromise<boolean>;
    cacheQueryResult(query: any, values: any, cacheKey: any, expiration: any, options: {
        renewalThreshold?: number;
        renewalKey?: any;
        priority?: number;
        external?: boolean;
        requestId?: string;
        dataSource: string;
        waitForRenew?: boolean;
        forceNoCache?: boolean;
        useInMemory?: boolean;
        useCsvQuery?: boolean;
    }): Promise<any>;
    protected lastRefreshTime(cacheKey: any): Promise<Date>;
    resultFromCacheIfExists(queryBody: any): Promise<{
        data: any;
        lastRefreshTime: Date;
    }>;
    queryRedisKey(cacheKey: any): string;
    cleanup(): Promise<void>;
    testConnection(): Promise<void>;
}
export {};
//# sourceMappingURL=QueryCache.d.ts.map