"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.QueryOrchestrator = exports.DriverType = void 0;
const ramda_1 = __importDefault(require("ramda"));
const shared_1 = require("@cubejs-backend/shared");
const QueryCache_1 = require("./QueryCache");
const PreAggregations_1 = require("./PreAggregations");
const RedisPool_1 = require("./RedisPool");
const RedisQueueEventsBus_1 = require("./RedisQueueEventsBus");
const LocalQueueEventsBus_1 = require("./LocalQueueEventsBus");
var DriverType;
(function (DriverType) {
    DriverType["External"] = "external";
    DriverType["Internal"] = "internal";
    DriverType["Cache"] = "cache";
})(DriverType = exports.DriverType || (exports.DriverType = {}));
class QueryOrchestrator {
    constructor(redisPrefix, driverFactory, logger, options = {}) {
        this.redisPrefix = redisPrefix;
        this.driverFactory = driverFactory;
        this.logger = logger;
        this.rollupOnlyMode = options.rollupOnlyMode;
        const cacheAndQueueDriver = options.cacheAndQueueDriver || shared_1.getEnv('cacheAndQueueDriver') || ((shared_1.getEnv('nodeEnv') === 'production' || shared_1.getEnv('redisUrl') || shared_1.getEnv('redisUseIORedis'))
            ? 'redis'
            : 'memory');
        this.cacheAndQueueDriver = cacheAndQueueDriver;
        if (!['redis', 'memory'].includes(cacheAndQueueDriver)) {
            throw new Error('Only \'redis\' or \'memory\' are supported for cacheAndQueueDriver option');
        }
        const redisPool = cacheAndQueueDriver === 'redis' ? new RedisPool_1.RedisPool(options.redisPoolOptions) : undefined;
        this.redisPool = redisPool;
        const { externalDriverFactory, continueWaitTimeout, skipExternalCacheAndQueue } = options;
        this.queryCache = new QueryCache_1.QueryCache(this.redisPrefix, driverFactory, this.logger, {
            externalDriverFactory,
            cacheAndQueueDriver,
            redisPool,
            continueWaitTimeout,
            skipExternalCacheAndQueue,
            ...options.queryCacheOptions,
        });
        this.preAggregations = new PreAggregations_1.PreAggregations(this.redisPrefix, this.driverFactory, this.logger, this.queryCache, {
            externalDriverFactory,
            cacheAndQueueDriver,
            redisPool,
            continueWaitTimeout,
            skipExternalCacheAndQueue,
            ...options.preAggregationsOptions,
            getQueueEventsBus: shared_1.getEnv('preAggregationsQueueEventsBus') &&
                this.getQueueEventsBus.bind(this)
        });
    }
    getQueueEventsBus() {
        if (!this.queueEventsBus) {
            const isRedis = this.cacheAndQueueDriver === 'redis';
            this.queueEventsBus = isRedis ?
                new RedisQueueEventsBus_1.RedisQueueEventsBus({ redisPool: this.redisPool }) :
                new LocalQueueEventsBus_1.LocalQueueEventsBus();
        }
        return this.queueEventsBus;
    }
    /**
     * Returns QueryCache instance.
     */
    getQueryCache() {
        return this.queryCache;
    }
    /**
     * Returns PreAggregations instance.
     */
    getPreAggregations() {
        return this.preAggregations;
    }
    /**
     * Force reconcile queue logic to be executed.
     */
    async forceReconcile(datasource = 'default') {
        await this.queryCache.forceReconcile(datasource);
    }
    async fetchQuery(queryBody) {
        var _a;
        const { preAggregationsTablesToTempTables, values } = await this.preAggregations.loadAllPreAggregationsIfNeeded(queryBody);
        if (values) {
            queryBody = {
                ...queryBody,
                values
            };
        }
        const usedPreAggregations = ramda_1.default.pipe(ramda_1.default.fromPairs, ramda_1.default.map((pa) => ({
            targetTableName: pa.targetTableName,
            refreshKeyValues: pa.refreshKeyValues,
            lastUpdatedAt: pa.lastUpdatedAt,
        })))(preAggregationsTablesToTempTables);
        if (this.rollupOnlyMode && Object.keys(usedPreAggregations).length === 0) {
            throw new Error('No pre-aggregation table has been built for this query yet. Please check your refresh worker configuration if it persists.');
        }
        let lastRefreshTimestamp = PreAggregations_1.getLastUpdatedAtTimestamp(preAggregationsTablesToTempTables.map(pa => new Date(pa[1].lastUpdatedAt)));
        if (!queryBody.query) {
            return {
                usedPreAggregations,
                lastRefreshTime: lastRefreshTimestamp && new Date(lastRefreshTimestamp),
            };
        }
        const result = await this.queryCache.cachedQueryResult(queryBody, preAggregationsTablesToTempTables);
        lastRefreshTimestamp = PreAggregations_1.getLastUpdatedAtTimestamp([lastRefreshTimestamp, (_a = result.lastRefreshTime) === null || _a === void 0 ? void 0 : _a.getTime()]);
        return {
            ...result,
            dataSource: queryBody.dataSource,
            external: queryBody.external,
            usedPreAggregations,
            lastRefreshTime: lastRefreshTimestamp && new Date(lastRefreshTimestamp),
        };
    }
    async loadRefreshKeys(query) {
        return this.queryCache.loadRefreshKeysFromQuery(query);
    }
    async queryStage(queryBody) {
        const preAggregationsQueryStageStateByDataSource = {};
        const preAggregationsQueryStageState = async (dataSource) => {
            if (!preAggregationsQueryStageStateByDataSource[dataSource]) {
                const queue = await this.preAggregations.getQueue(dataSource);
                preAggregationsQueryStageStateByDataSource[dataSource] = queue.fetchQueryStageState();
            }
            return preAggregationsQueryStageStateByDataSource[dataSource];
        };
        const pendingPreAggregationIndex = (await Promise.all((queryBody.preAggregations || [])
            .map(async (p) => {
            const queue = await this.preAggregations.getQueue(p.dataSource);
            return queue.getQueryStage(PreAggregations_1.PreAggregations.preAggregationQueryCacheKey(p), 10, await preAggregationsQueryStageState(p.dataSource));
        }))).findIndex(p => !!p);
        if (pendingPreAggregationIndex === -1) {
            const qcQueue = await this.queryCache.getQueue(queryBody.dataSource);
            return qcQueue.getQueryStage(QueryCache_1.QueryCache.queryCacheKey(queryBody));
        }
        const preAggregation = queryBody.preAggregations[pendingPreAggregationIndex];
        const paQueue = await this.preAggregations.getQueue(preAggregation.dataSource);
        const preAggregationStage = await paQueue.getQueryStage(PreAggregations_1.PreAggregations.preAggregationQueryCacheKey(preAggregation), undefined, await preAggregationsQueryStageState(preAggregation.dataSource));
        if (!preAggregationStage) {
            return undefined;
        }
        const stageMessage = `Building pre-aggregation ${pendingPreAggregationIndex + 1}/${queryBody.preAggregations.length}`;
        if (preAggregationStage.stage.indexOf('queue') !== -1) {
            return { ...preAggregationStage, stage: `${stageMessage}: ${preAggregationStage.stage}` };
        }
        else {
            return { ...preAggregationStage, stage: stageMessage };
        }
    }
    resultFromCacheIfExists(queryBody) {
        return this.queryCache.resultFromCacheIfExists(queryBody);
    }
    async testConnections() {
        // @todo Possible, We will allow to use different drivers for cache and queue, dont forget to add both
        try {
            await this.queryCache.testConnection();
        }
        catch (e) {
            e.driverType = DriverType.Cache;
            throw e;
        }
    }
    async cleanup() {
        return this.queryCache.cleanup();
    }
    async getPreAggregationVersionEntries(preAggregations, preAggregationsSchema, requestId) {
        const versionEntries = await this.preAggregations.getVersionEntries(preAggregations.map(p => {
            const { preAggregation } = p.preAggregation;
            const partition = p.partitions[0];
            preAggregation.dataSource = (partition && partition.dataSource) || 'default';
            preAggregation.preAggregationsSchema = preAggregationsSchema;
            return preAggregation;
        }), requestId);
        const flatFn = (arrResult, arrItem) => ([...arrResult, ...arrItem]);
        const structureVersionsByTableName = preAggregations
            .map(p => p.partitions)
            .reduce(flatFn, [])
            .reduce((obj, partition) => {
            if (partition) {
                obj[partition.tableName] = PreAggregations_1.PreAggregations.structureVersion(partition);
            }
            return obj;
        }, {});
        return {
            structureVersionsByTableName,
            versionEntriesByTableName: versionEntries
                .reduce(flatFn, [])
                .filter((versionEntry) => {
                const structureVersion = structureVersionsByTableName[versionEntry.table_name];
                return structureVersion && versionEntry.structure_version === structureVersion;
            })
                .reduce((obj, versionEntry) => {
                if (!obj[versionEntry.table_name])
                    obj[versionEntry.table_name] = [];
                obj[versionEntry.table_name].push(versionEntry);
                return obj;
            }, {})
        };
    }
    async getPreAggregationPreview(requestId, preAggregation) {
        if (!preAggregation)
            return [];
        const [query] = preAggregation.previewSql;
        const { external } = preAggregation;
        const data = await this.fetchQuery({
            continueWait: true,
            query,
            external,
            preAggregations: [
                preAggregation
            ],
            requestId,
        });
        return data || [];
    }
    async expandPartitionsInPreAggregations(queryBody) {
        return this.preAggregations.expandPartitionsInPreAggregations(queryBody);
    }
    async checkPartitionsBuildRangeCache(queryBody) {
        return this.preAggregations.checkPartitionsBuildRangeCache(queryBody);
    }
    async getPreAggregationQueueStates(dataSource = 'default') {
        return this.preAggregations.getQueueState(dataSource);
    }
    async cancelPreAggregationQueriesFromQueue(queryKeys, dataSource = 'default') {
        return this.preAggregations.cancelQueriesFromQueue(queryKeys, dataSource);
    }
    async subscribeQueueEvents(id, callback) {
        return this.getQueueEventsBus().subscribe(id, callback);
    }
    async unSubscribeQueueEvents(id) {
        return this.getQueueEventsBus().unsubscribe(id);
    }
}
exports.QueryOrchestrator = QueryOrchestrator;
//# sourceMappingURL=QueryOrchestrator.js.map